from secrets import token_bytes

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import fields
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from nuntius import app_settings


class CampaignStatusType:
    STATUS_WAITING = 0
    STATUS_SENDING = 1
    STATUS_SENT = 2
    STATUS_ERROR = 3
    STATUS_CHOICES = (
        (STATUS_WAITING, _("Waiting")),
        (STATUS_SENDING, _("Sending")),
        (STATUS_SENT, _("Sent")),
        (STATUS_ERROR, _("Error")),
    )


class AbstractCampaignQuerySet(models.QuerySet):
    def outbox(self):
        now = timezone.now()
        return (
            self.filter(status__lt=CampaignStatusType.STATUS_SENT)
            .exclude(start_date__isnull=False, start_date__gt=now)
            .exclude(end_date__isnull=False, end_date__lt=now)
        )


class AbstractCampaign(CampaignStatusType, models.Model):
    objects = AbstractCampaignQuerySet.as_manager()

    name = fields.CharField(_("Name (invisible to subscribers)"), max_length=255)
    created = fields.DateTimeField(_("Created"), auto_now_add=True)
    updated = fields.DateTimeField(_("Updated"), auto_now=True)
    first_sent = fields.DateTimeField(_("First sent"), blank=True, null=True)

    segment = models.ForeignKey(
        to=app_settings.NUNTIUS_SEGMENT_MODEL,
        verbose_name=_("Subscriber segment"),
        on_delete=models.SET_NULL,
        null=True,
    )
    start_date = models.DateTimeField(
        verbose_name=_("Campaign start date"), null=True, blank=True
    )
    end_date = models.DateTimeField(
        verbose_name=_("Campaign end date"), null=True, blank=True
    )

    status = fields.IntegerField(
        choices=CampaignStatusType.STATUS_CHOICES,
        default=CampaignStatusType.STATUS_WAITING,
    )

    utm_name = fields.CharField(
        _("UTM name (visible to subscribers)"),
        max_length=255,
        blank=True,
        help_text=_(
            "Value used as utm_campaign parameter, used by various analytics tools."
        ),
    )

    def generate_signature_key():
        return token_bytes(20)

    signature_key = fields.BinaryField(max_length=20, default=generate_signature_key)

    def get_subscribers_queryset(self):
        if self.segment is None:
            model = app_settings.NUNTIUS_SUBSCRIBER_MODEL
            model_class = ContentType.objects.get(
                app_label=model.split(".")[0], model=model.split(".")[1].lower()
            ).model_class()
            return model_class.objects.all()
        else:
            return self.segment.get_subscribers_queryset()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"AbstractCampaign(id={self.id!r}, name={self.name!r})"

    class Meta:
        abstract = True
        verbose_name = _("Abstract campaign")
        verbose_name_plural = _("Abstract campaigns")



# Generated by Django 2.2.23 on 2022-03-01 08:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0022_add_start_and_end_dates_to_campaigns")]

    operations = [
        migrations.AlterField(
            model_name="pushcampaign",
            name="notification_tag",
            field=models.CharField(
                blank=True,
                help_text="Tagging a notification allows replacing it by pushing another with the same tag value",
                max_length=255,
                null=True,
                verbose_name="Notification tag",
            ),
        )
    ]



from urllib.parse import quote as url_quote

from django.urls import reverse

from nuntius import app_settings
from nuntius.models import PushCampaignSentStatusType
from nuntius.utils.messages import sign_url, extend_query

from firebase_admin import messaging

def get_pushing_error_classes():
    try:
        from push_notifications import gcm
    except ImportError:
        classes = tuple()
    else:
        classes = (gcm.FirebaseError)

    return classes


def make_tracking_url(url, campaign, tracking_id):
    url = extend_query(
        url, defaults={"utm_term": getattr(campaign.segment, "utm_term", "")}
    )

    relative_url = reverse(
        "nuntius_track_push_click",
        kwargs={
            "tracking_id": tracking_id,
            "signature": sign_url(campaign, url),
            "link": url_quote(url, safe=""),
        },
    )
    return f"{app_settings.LINKS_URL}{relative_url}"


def notification_for_event(sent_event):
    """Generate a push notification payload corresponding to a PushCampaignSentEvent instance

    :param sent_event: the PushCampaignSentEvent instance associating a campaign and a subscriber for which the
        notification should be generated
    :type sent_event: class:`nuntius.models.PushCampaignSentEvent`
    :return: a dictionnary describing the push notification content
    :rtype: dict`
    """

    # subscriber = sent_event.subscriber
    campaign = sent_event.campaign
    notification_url = make_tracking_url(
        campaign.notification_url, campaign, tracking_id=sent_event.tracking_id
    )
    notification = {
        "title": campaign.notification_title,
        "url": notification_url,
        "body": campaign.notification_body,
        "tag": campaign.notification_tag,
        "icon": campaign.notification_icon,
    }

    return notification

def push_gcm_notification(device, notification, thread_id):
    ttl = 259200 # equals 3 days
    push_message = messaging.Message(
        data={
            "url": notification["url"]
        },
        notification=messaging.Notification(
            title=notification["title"],
            body=notification["body"],
            image=notification["notification_icon"]
        ),
        android=messaging.AndroidConfig(
            ttl=ttl,
            collapse_key=thread_id
        ),
        apns=messaging.APNSConfig(
            headers={
                "apns-expiration": ttl,
                "apns-collapse-id": thread_id
            }
        )
    )
    device.send_message(push_message)


def push_notification(notification, push_sent_event):
    try:
        from push_notifications.models import GCMDevice
    except ImportError:
        push_sent_event.result = PushCampaignSentStatusType.ERROR
    else:
        push_sent_event.result = PushCampaignSentStatusType.PENDING

        pushed_count = 0
        for device in push_sent_event.devices:
            try:
                if isinstance(device, GCMDevice):
                    push_gcm_notification(
                        device, notification, push_sent_event.campaign.id
                    )
            except Exception:
                pass
            else:
                pushed_count += 1

        if pushed_count > 0:
            push_sent_event.result = PushCampaignSentStatusType.OK
        else:
            push_sent_event.result = PushCampaignSentStatusType.ERROR

    push_sent_event.save()



from secrets import token_urlsafe

from django.db import models, IntegrityError
from django.db.models import fields
from django.utils.translation import gettext_lazy as _

from nuntius import app_settings
from nuntius.models import AbstractCampaign


class PushCampaign(AbstractCampaign):
    notification_title = fields.CharField(_("Notification title"), max_length=255)
    notification_url = fields.CharField(_("Notification URL"), max_length=255)
    notification_body = fields.TextField(_("Notification body"))
    notification_tag = fields.CharField(
        _("Notification tag"),
        null=True,
        blank=True,
        max_length=255,
        help_text=_(
            "Tagging a notification allows replacing it by pushing another with the same tag value"
        ),
    )
    notification_icon = fields.CharField(
        _("Notification icon"), null=True, blank=True, max_length=255
    )

    def get_sent_count(self):
        return (
            PushCampaignSentEvent.objects.filter(campaign=self)
            .exclude(result=PushCampaignSentStatusType.PENDING)
            .count()
        )

    def get_ok_count(self):
        return (
            PushCampaignSentEvent.objects.filter(campaign=self)
            .filter(result=PushCampaignSentStatusType.OK)
            .count()
        )

    def get_ko_count(self):
        return (
            PushCampaignSentEvent.objects.filter(campaign=self)
            .filter(result=PushCampaignSentStatusType.ERROR)
            .count()
        )

    def get_click_count(self):
        return (
            PushCampaignSentEvent.objects.filter(campaign=self)
            .filter(click_count__gt=0)
            .count()
        )

    def get_event_for_subscriber(self, subscriber):
        event, _ = PushCampaignSentEvent.objects.get_or_create(
            campaign=self, subscriber=subscriber
        )
        return event

    def __repr__(self):
        return f"PushCampaign(id={self.id!r}, name={self.name!r})"

    class Meta:
        verbose_name = _("push campaign")
        verbose_name_plural = _("push campaigns")


class PushCampaignSentStatusType:
    UNKNOWN = "?"
    PENDING = "P"
    OK = "OK"
    ERROR = "E"

    CHOICES = (
        (UNKNOWN, _("Unknown")),
        (PENDING, _("Sending")),
        (OK, _("Sent")),
        (ERROR, _("Error")),
    )


class PushCampaignSentEvent(models.Model):
    subscriber = models.ForeignKey(
        app_settings.NUNTIUS_SUBSCRIBER_MODEL,
        models.SET_NULL,
        verbose_name=_("Subscriber"),
        null=True,
        blank=True,
    )
    campaign = models.ForeignKey(
        "PushCampaign", models.CASCADE, verbose_name=_("Push campaign")
    )
    datetime = models.DateTimeField(_("Sending time"), auto_now_add=True)
    result = models.CharField(
        _("Operation result"),
        max_length=2,
        default=PushCampaignSentStatusType.PENDING,
        choices=PushCampaignSentStatusType.CHOICES,
    )

    def generate_tracking_id():
        return token_urlsafe(9)

    tracking_id = models.CharField(
        max_length=12,
        default=generate_tracking_id,
        null=True,
        editable=False,
        unique=True,
    )

    click_count = models.IntegerField(_("Click count"), default=0, editable=False)

    @property
    def devices(self):
        return self.subscriber.get_subscriber_push_devices()

    class Meta:
        verbose_name = _("push sent event")
        verbose_name_plural = _("push sent events")
        constraints = [
            models.UniqueConstraint(
                name="unique_push_campaign_subscriber",
                fields=["campaign", "subscriber"],
            )
        ]
        indexes = [models.Index(fields=["subscriber", "datetime"])]
        ordering = ["-datetime"]



# Generated by Django 2.2.23 on 2021-11-09 05:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import nuntius.models.mixins
import nuntius.models.push_campaigns


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.NUNTIUS_SEGMENT_MODEL),
        migrations.swappable_dependency(settings.NUNTIUS_SUBSCRIBER_MODEL),
        ("nuntius", "0020_auto_20210118_1100"),
    ]

    operations = [
        migrations.CreateModel(
            name="PushCampaign",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=255, verbose_name="Name (invisible to subscribers)"
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created"),
                ),
                (
                    "updated",
                    models.DateTimeField(auto_now=True, verbose_name="Updated"),
                ),
                (
                    "first_sent",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="First sent"
                    ),
                ),
                (
                    "status",
                    models.IntegerField(
                        choices=[
                            (0, "Waiting"),
                            (1, "Sending"),
                            (2, "Sent"),
                            (3, "Error"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "utm_name",
                    models.CharField(
                        blank=True,
                        help_text="Value used as utm_campaign parameter, used by various analytics tools.",
                        max_length=255,
                        verbose_name="UTM name (visible to subscribers)",
                    ),
                ),
                (
                    "signature_key",
                    models.BinaryField(
                        default=nuntius.models.mixins.AbstractCampaign.generate_signature_key,
                        max_length=20,
                    ),
                ),
                (
                    "notification_title",
                    models.CharField(max_length=255, verbose_name="Notification title"),
                ),
                (
                    "notification_url",
                    models.CharField(max_length=255, verbose_name="Notification URL"),
                ),
                (
                    "notification_body",
                    models.TextField(verbose_name="Notification body"),
                ),
                (
                    "notification_tag",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Notification tag",
                    ),
                ),
                (
                    "notification_icon",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Notification icon",
                    ),
                ),
                (
                    "segment",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.NUNTIUS_SEGMENT_MODEL,
                        verbose_name="Subscriber segment",
                    ),
                ),
            ],
            options={
                "verbose_name": "push campaign",
                "verbose_name_plural": "push campaigns",
            },
            bases=(nuntius.models.mixins.CampaignStatusType, models.Model),
        ),
        migrations.AlterModelOptions(
            name="campaign",
            options={
                "verbose_name": "email campaign",
                "verbose_name_plural": "email campaigns",
            },
        ),
        migrations.AlterModelOptions(
            name="campaignsentevent",
            options={
                "ordering": ["-datetime"],
                "verbose_name": "email sent event",
                "verbose_name_plural": "email sent events",
            },
        ),
        migrations.CreateModel(
            name="PushCampaignSentEvent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "datetime",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Sending time"
                    ),
                ),
                (
                    "result",
                    models.CharField(
                        choices=[
                            ("?", "Unknown"),
                            ("P", "Sending"),
                            ("OK", "Sent"),
                            ("E", "Error"),
                        ],
                        default="P",
                        max_length=2,
                        verbose_name="Operation result",
                    ),
                ),
                (
                    "tracking_id",
                    models.CharField(
                        default=nuntius.models.push_campaigns.PushCampaignSentEvent.generate_tracking_id,
                        editable=False,
                        max_length=12,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "click_count",
                    models.IntegerField(
                        default=0, editable=False, verbose_name="Click count"
                    ),
                ),
                (
                    "campaign",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="nuntius.PushCampaign",
                        verbose_name="Push campaign",
                    ),
                ),
                (
                    "subscriber",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.NUNTIUS_SUBSCRIBER_MODEL,
                        verbose_name="Subscriber",
                    ),
                ),
            ],
            options={
                "verbose_name": "push sent event",
                "verbose_name_plural": "push sent events",
                "ordering": ["-datetime"],
            },
        ),
        migrations.AddIndex(
            model_name="pushcampaignsentevent",
            index=models.Index(
                fields=["subscriber", "datetime"], name="nuntius_pus_subscri_9ce7a8_idx"
            ),
        ),
        migrations.AddConstraint(
            model_name="pushcampaignsentevent",
            constraint=models.UniqueConstraint(
                fields=("campaign", "subscriber"),
                name="unique_push_campaign_subscriber",
            ),
        ),
    ]



# Generated by Django 3.2.18 on 2023-03-24 11:16

from django.db import migrations
import stdimage.models


class Migration(migrations.Migration):
    dependencies = [
        ("nuntius", "0023_add_help_text_to_pushcampaign_notification_tag"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mosaicoimage",
            name="file",
            field=stdimage.models.StdImageField(
                force_min_size=False,
                upload_to="mosaico",
                variations={"thumbnail": (90, 90)},
            ),
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _
from django.conf import settings

CAMPAIGN_TYPE_EMAIL = "email"
CAMPAIGN_TYPE_PUSH = "push"
ENABLED_CAMPAIGN_TYPES = getattr(
    settings, "NUNTIUS_ENABLED_CAMPAIGN_TYPES", [CAMPAIGN_TYPE_EMAIL]
)

# Segment and subscribers models
try:
    NUNTIUS_SEGMENT_MODEL = settings.NUNTIUS_SEGMENT_MODEL
    NUNTIUS_SUBSCRIBER_MODEL = settings.NUNTIUS_SUBSCRIBER_MODEL
except AttributeError:
    raise ImproperlyConfigured(
        "Both NUNTIUS_SEGMENT_MODEL and NUNTIUS_SUBSCRIBER_MODEL must be set to valid and existing models. See documentation for more information."
    )

MOSAICO_TEMPLATES = getattr(
    settings,
    "NUNTIUS_MOSAICO_TEMPLATES",
    [
        (
            f"{settings.STATIC_URL}/nuntius/templates/versafix-1/template-versafix-1.html",
            _("Default template"),
        )
    ],
)

DEFAULT_FROM_EMAIL = getattr(settings, "NUNTIUS_DEFAULT_FROM_EMAIL", "")
DEFAULT_FROM_NAME = getattr(settings, "NUNTIUS_DEFAULT_FROM_NAME", "")
DEFAULT_REPLY_TO_EMAIL = getattr(settings, "NUNTIUS_DEFAULT_REPLY_TO_EMAIL", "")
DEFAULT_REPLY_TO_NAME = getattr(settings, "NUNTIUS_DEFAULT_REPLY_TO_NAME", "")

DISABLE_DEFAULT_ADMIN = getattr(settings, "NUNTIUS_DISABLE_DEFAULT_ADMIN", False)

PUBLIC_URL = getattr(settings, "NUNTIUS_PUBLIC_URL", None)
IMAGES_URL = getattr(settings, "NUNTIUS_IMAGES_URL", PUBLIC_URL)
LINKS_URL = getattr(settings, "NUNTIUS_LINKS_URL", PUBLIC_URL)

BOUNCE_PARAMS = {
    "consecutive": 1,
    "duration": 7,
    "limit": 3,
    **getattr(settings, "NUNTIUS_BOUNCE_PARAMS", {}),
}
EMAIL_BACKEND = getattr(settings, "NUNTIUS_EMAIL_BACKEND", None)

# Maximum rate with which emails may be sent, by number of emails per second
MAX_SENDING_RATE = getattr(settings, "NUNTIUS_MAX_SENDING_RATE", 50)

# Number of concurrent sending processes to send emails
MAX_CONCURRENT_SENDERS = getattr(settings, "NUNTIUS_MAX_CONCURRENT_SENDERS", 4)

# Maximum number of messages that may be sent over a single SMTP connection
MAX_MESSAGES_PER_CONNECTION = getattr(
    settings, "NUNTIUS_MAX_MESSAGES_PER_SMTP_CONNECTION", 500
)

# Interval of time, in seconds, with which the worker must check for campaign status changes
POLLING_INTERVAL = getattr(settings, "NUNTIUS_POLLING_INTERVAL", 2)

if CAMPAIGN_TYPE_PUSH in ENABLED_CAMPAIGN_TYPES:
    try:
        PUSH_NOTIFICATION_SETTINGS = settings.NUNTIUS_PUSH_NOTIFICATION_SETTINGS
    except AttributeError:
        raise ImproperlyConfigured(
            "Push campaigns have been enabled without specifying a configuration. Either remove push from the list of "
            "enabled campaign, or add a valid 'django-push-notification' configuration ("
            "settings.NUNTIUS_PUSH_NOTIFICATION_SETTINGS)."
        )



import contextlib
import logging
import multiprocessing as mp
import multiprocessing.connection as mpc
import signal
import smtplib
from argparse import ArgumentTypeError
from typing import Dict, List, Tuple

from django.core import mail
from django.core.mail import EmailMessage
from django.core.management import BaseCommand
from django.db import connection
from django.db.models import Exists, OuterRef
from django.utils import timezone
from django.utils.translation import gettext as _, gettext_lazy
from tenacity import (
    retry,
    stop_after_attempt,
    retry_if_exception_type,
    wait_random_exponential,
    TryAgain,
)

from nuntius import app_settings
from nuntius.app_settings import CAMPAIGN_TYPE_EMAIL, CAMPAIGN_TYPE_PUSH
from nuntius.messages import message_for_event
from nuntius.models import (
    Campaign,
    CampaignSentEvent,
    CampaignSentStatusType,
    PushCampaign,
    PushCampaignSentEvent,
    PushCampaignSentStatusType,
    AbstractSubscriber,
)
from nuntius.utils.notifications import (
    notification_for_event,
    push_notification,
    get_pushing_error_classes,
)
from nuntius.utils.processes import (
    gracefully_exit,
    print_stack_trace,
    reset_sigmask,
    RateLimiter,
    TokenBucket,
    RateMeter,
    GracefulExit,
    get_from_queue_or_quit,
    put_in_queue_or_quit,
    unexpected_exc_logger,
)

try:
    from anymail import exceptions as anymail_exceptions

    AnymailError = anymail_exceptions.AnymailError
    AnymailAPIError = anymail_exceptions.AnymailAPIError
    AnymailRecipientsRefused = anymail_exceptions.AnymailRecipientsRefused
except:

    class FakeException(Exception):
        pass

    AnymailError = AnymailRecipientsRefused = AnymailAPIError = FakeException


logger = logging.getLogger(__name__)


class ConnectionManager:
    """
    Manager around the SMTP or Mail API connection that handles reconnection and quitting
    """

    CONNECTION_ERRORS = (ConnectionError, smtplib.SMTPException, AnymailError)
    MAIL_SENDING_ERRORS = (
        smtplib.SMTPSenderRefused,
        smtplib.SMTPDataError,
        AnymailAPIError,
    )

    def __init__(self, quit_event):
        self._connection = mail.get_connection(backend=app_settings.EMAIL_BACKEND)
        self._message_counter = 0
        self._quit_event = quit_event

    @retry(
        wait=wait_random_exponential(max=30),
        retry=retry_if_exception_type(CONNECTION_ERRORS),
    )
    def open_connection(self):
        """
        Connect to the SMTP or API server and retry with exponential backoff.

        This function will try up to 5 times to connect to the server, allowing for shaky
        connections.
        """
        if self._quit_event.is_set():
            raise GracefulExit()

        try:
            self._connection.open()
        except Exception:
            self._connection.close()
            raise
        else:
            self._message_counter = 0

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_random_exponential(),
        retry=retry_if_exception_type(MAIL_SENDING_ERRORS),
    )
    def send_message(self, message: EmailMessage):
        """
        Send an email message and retry in cases of failures.

        This function will try sending up to 5 times, and uses an exponential
        backoff strategy, where a random waiting time is chosen between 0 and a max
        value that doubles every retry, to avoid all sender processes retrying at the
        same time.

        :param message: the email message to send
        :param connection: the connection to use to send the message
        """
        if self._quit_event.is_set():
            raise GracefulExit()

        if (
            app_settings.MAX_MESSAGES_PER_CONNECTION
            and self._message_counter >= app_settings.MAX_MESSAGES_PER_CONNECTION
        ):
            self._connection.close()
            self.open_connection()
        try:
            self._connection.send_messages([message])
            self._message_counter += 1
        except smtplib.SMTPServerDisconnected:
            self._connection.close()
            self.open_connection()
            raise TryAgain

    def __enter__(self):
        self.open_connection()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._connection.close()


class PushManager:
    """
    Manager around the push notification sending that handles retrying
    """

    def __init__(self, quit_event):
        self._quit_event = quit_event

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_random_exponential(),
        retry=retry_if_exception_type(get_pushing_error_classes()),
    )
    def push(self, notification, push_sent_event):
        """
        Send a push notification and retry in cases of failures.

        This function will try sending up to 5 times, and uses an exponential
        backoff strategy, where a random waiting time is chosen between 0 and a max
        value that doubles every retry, to avoid all sender processes retrying at the
        same time.
        """
        if self._quit_event.is_set():
            raise GracefulExit()

        push_notification(notification, push_sent_event)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


@reset_sigmask
@unexpected_exc_logger
def mailer_process(
    *,
    queue: mp.Queue,
    error_channel: mpc.Connection,
    quit_event: mp.Event,
    rate_limiter: RateLimiter = None,
    rate_meter: RateMeter = None,
):
    """
    Main function of the processes responsible for sending email messages.

    This process pulls `(message: EmailMessage, event: CampaignSentEvent)`
    tuples from the work queue, tries to send the message and saves the result
    on the `event`.

    Whenever an unexpected error happens (i.e. which does not seem to be linked
    to a particular recipient), the process signals the error on the
    `error_channel` by sending the campaign id.

    It monitors its own connection to the mail service, and resets it every
    `nuntius.app_settings.MAX_MESSAGES_PER_CONNECTION` messages.

    :param queue: The work queue on which tuples (EmailMessage, CampaignSentEvent) are received.
    :type queue: class:`multiprocessing.Queue`

    :param error_channel: channel on which campaign ids of failing campaigns are sent.
    :type error_channel: class:`multiprocessing.connection.Connection`

    :param quit_event: event that may be used by the main process to signal to senders they need to quit
    :type quit_event: class:`multiprocessing.Event`

    :param rate_limiter: rate limiter to limit the rate of message sendings.
    :type rate_limiter: class:`nuntius.utils.processes.RateLimiter`

    :param rate_meter: rate meter to allow measuring the sending speed
    :type rate_meter: class:`nuntius.utils.processes.RateMeter`
    """
    message: EmailMessage
    sent_event_id: int

    try:
        with ConnectionManager(quit_event) as connection_manager:
            while True:
                # the timeout allows the loop to start again every few seconds so that the
                # quit_event is checked and the process can quit if it has to.
                message, sent_event_id = get_from_queue_or_quit(
                    queue,
                    event=quit_event,
                    polling_period=app_settings.POLLING_INTERVAL,
                )

                sent_event_qs = CampaignSentEvent.objects.filter(id=sent_event_id)
                # messages generated by the campaign manager should only have one recipient
                email = message.to[0]

                # rate limit just before sending
                if rate_limiter:
                    rate_limiter.take()

                try:
                    connection_manager.send_message(message)
                except (smtplib.SMTPRecipientsRefused, AnymailRecipientsRefused):
                    # exceptions linked to a specific recipient need not stop the sending
                    CampaignSentEvent.objects.filter(id=sent_event_id).update(
                        result=CampaignSentStatusType.BLOCKED
                    )
                except GracefulExit:
                    raise
                except Exception:
                    campaign = Campaign.objects.get(campaignsentevent__id=sent_event_id)
                    error_channel.send(campaign.id)
                    logger.error(
                        _("Error while sending email for campaign %(campaign)s")
                        % repr(campaign),
                        exc_info=True,
                    )

                else:
                    if rate_meter:
                        rate_meter.count_up()
                    if hasattr(message, "anymail_status"):
                        if message.anymail_status.recipients[email].status in [
                            "invalid",
                            "rejected",
                            "failed",
                        ]:
                            sent_event_qs.update(result=CampaignSentStatusType.REJECTED)
                        else:
                            sent_event_qs.update(
                                result=CampaignSentStatusType.UNKNOWN,
                                esp_message_id=message.anymail_status.recipients[
                                    email
                                ].message_id,
                            )
                    else:
                        sent_event_qs.update(result=CampaignSentStatusType.UNKNOWN)
    except GracefulExit:
        return


@reset_sigmask
@unexpected_exc_logger
def pusher_process(
    *,
    queue: mp.Queue,
    error_channel: mpc.Connection,
    quit_event: mp.Event,
    rate_limiter: RateLimiter = None,
    rate_meter: RateMeter = None,
):
    """
    Main function of the processes responsible for sending push notifications.

    This process pulls `(push_notification, event: PushCampaignSentEvent)`
    tuples from the work queue, tries to send the push and saves the result
    on the `event`.

    Whenever an unexpected error happens (i.e. which does not seem to be linked
    to a particular recipient), the process signals the error on the
    `error_channel` by sending the campaign id.

    It monitors its own connection to the mail service, and resets it every
    `nuntius.app_settings.MAX_MESSAGES_PER_CONNECTION` messages.

    :param queue: The work queue on which tuples (push_notification, CampaignSentEvent) are received.
    :type queue: class:`multiprocessing.Queue`

    :param error_channel: channel on which campaign ids of failing campaigns are sent.
    :type error_channel: class:`multiprocessing.connection.Connection`

    :param quit_event: event that may be used by the main process to signal to senders they need to quit
    :type quit_event: class:`multiprocessing.Event`

    :param rate_limiter: rate limiter to limit the rate of message sendings.
    :type rate_limiter: class:`nuntius.utils.processes.RateLimiter`

    :param rate_meter: rate meter to allow measuring the sending speed
    :type rate_meter: class:`nuntius.utils.processes.RateMeter`
    """
    notification: dict
    push_sent_event_id: int

    try:
        with PushManager(quit_event) as push_manager:
            while True:
                # the timeout allows the loop to start again every few seconds so that the
                # quit_event is checked and the process can quit if it has to.
                notification, push_sent_event_id = get_from_queue_or_quit(
                    queue,
                    event=quit_event,
                    polling_period=app_settings.POLLING_INTERVAL,
                )

                # rate limit just before sending
                if rate_limiter:
                    rate_limiter.take()

                try:
                    push_sent_event = PushCampaignSentEvent.objects.get(
                        id=push_sent_event_id
                    )
                    push_manager.push(notification, push_sent_event)
                except GracefulExit:
                    raise
                except PushCampaignSentEvent.DoesNotExist:
                    logger.error(
                        _(
                            f"Push campaign sent event with id '{push_sent_event_id}' not found"
                        ),
                        exc_info=True,
                    )
                except Exception:
                    push_campaign = PushCampaign.objects.get(
                        pushcampaignsentevent__id=push_sent_event_id
                    )
                    error_channel.send(push_campaign.id)
                    logger.error(
                        _(
                            f"Error while pushing notification for campaign {repr(push_campaign)}"
                        ),
                        exc_info=True,
                    )
                else:
                    if rate_meter:
                        rate_meter.count_up()

    except GracefulExit:
        return


@reset_sigmask
@unexpected_exc_logger
def email_campaign_manager_process(
    *, campaign: Campaign, queue: mp.Queue, quit_event: mp.Event
):
    """
    Main function of the process responsible for scheduling the sending of campaigns

    :param campaign: the campaign for which messages must be sent
    :type campaign: :class:`nuntius.models.Campaign`

    :param queue: the work queue on which email messages are put
    :type queue: : class:`multiprocessing.Queue`

    :param quit_event: an event that may be used by the main process to tell the manager it needs to quit
    :type quit_event: class:`multiprocessing.Event`
    """
    queryset = campaign.get_subscribers_queryset()
    # eliminate people who already received the message
    queryset = queryset.annotate(
        already_sent=Exists(
            CampaignSentEvent.objects.filter(
                subscriber_id=OuterRef("pk"), campaign_id=campaign.id
            ).exclude(result=CampaignSentStatusType.PENDING)
        )
    ).filter(already_sent=False)

    campaign_finished = False

    for subscriber in queryset.iterator():
        if quit_event.is_set():
            break

        if subscriber.get_subscriber_status() != AbstractSubscriber.STATUS_SUBSCRIBED:
            continue

        sent_event = campaign.get_event_for_subscriber(subscriber)

        # just in case there is another nuntius_worker started, but this should not happen
        if sent_event.result != CampaignSentStatusType.PENDING:
            continue

        message = message_for_event(sent_event)

        try:
            put_in_queue_or_quit(
                queue,
                (message, sent_event.id),
                event=quit_event,
                polling_period=app_settings.POLLING_INTERVAL,
            )
        except GracefulExit:
            break
    else:
        campaign_finished = True

    queue.close()
    queue.join_thread()
    # everything has been scheduled for sending:
    if campaign_finished:
        campaign.status = Campaign.STATUS_SENT
        campaign.save()


@reset_sigmask
@unexpected_exc_logger
def push_campaign_manager_process(
    *, campaign: PushCampaign, queue: mp.Queue, quit_event: mp.Event
):
    """
    Main function of the process responsible for scheduling the pushing of campaigns

    :param campaign: the campaign for which messages must be pushed
    :type campaign: :class:`nuntius.models.PushCampaign`

    :param queue: the work queue on which email messages are put
    :type queue: : class:`multiprocessing.Queue`

    :param quit_event: an event that may be used by the main process to tell the manager it needs to quit
    :type quit_event: class:`multiprocessing.Event`
    """
    queryset = campaign.get_subscribers_queryset()
    # eliminate people who already received the message
    queryset = queryset.annotate(
        already_sent=Exists(
            PushCampaignSentEvent.objects.filter(
                subscriber_id=OuterRef("pk"), campaign_id=campaign.id
            ).exclude(result=PushCampaignSentStatusType.PENDING)
        )
    ).filter(already_sent=False)

    campaign_finished = False

    for subscriber in queryset.iterator():
        if quit_event.is_set():
            break

        if subscriber.get_subscriber_status() != AbstractSubscriber.STATUS_SUBSCRIBED:
            continue

        push_sent_event = campaign.get_event_for_subscriber(subscriber)

        if (
            push_sent_event is None
            or push_sent_event.result != PushCampaignSentStatusType.PENDING
        ):
            # just in case there is another nuntius_worker started, but this should not happen
            continue

        notification = notification_for_event(push_sent_event)

        try:
            put_in_queue_or_quit(
                queue,
                (notification, push_sent_event.id),
                event=quit_event,
                polling_period=app_settings.POLLING_INTERVAL,
            )
        except GracefulExit:
            break
    else:
        campaign_finished = True

    queue.close()
    queue.join_thread()
    # everything has been scheduled for sending:
    if campaign_finished:
        campaign.status = Campaign.STATUS_SENT
        if campaign.first_sent is None:
            campaign.first_sent = timezone.now()
        campaign.save()


CAMPAIGN_TYPE = {
    CAMPAIGN_TYPE_EMAIL: {
        "CampaignModel": Campaign,
        "sender_process": mailer_process,
        "manager_process": email_campaign_manager_process,
    },
    CAMPAIGN_TYPE_PUSH: {
        "CampaignModel": PushCampaign,
        "sender_process": pusher_process,
        "manager_process": push_campaign_manager_process,
    },
}


def campaign_types_argument(campaign_types):
    if not isinstance(campaign_types, str):
        raise ArgumentTypeError("The campaign type must be a string")

    campaign_types = campaign_types.lower().split(",")

    for campaign_type in campaign_types:
        if campaign_type not in CAMPAIGN_TYPE.keys():
            raise ArgumentTypeError(
                f"'{campaign_type}': this campaign type does not exist, possible choices are: {', '.join(CAMPAIGN_TYPE.keys())}"
            )

    return campaign_types


class Command(BaseCommand):
    STATS_MESSAGE = gettext_lazy(
        """
%(campaign_type)s STATISTICS
Message queue size: %(queue_size)s
Sender processes: %(sender_processes)s
Campaign managers: %(campaign_managers)s
Token bucket current capacity: %(bucket_capacity)s
Current sending rate: %(sending_rate)s"""
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-t",
            "--types",
            dest="campaign_types",
            default=app_settings.ENABLED_CAMPAIGN_TYPES,
            type=campaign_types_argument,
            help="The comma-separated list of campaign types for which to start the workers. "
            f"Valid choices are: {', '.join(CAMPAIGN_TYPE.keys())}. "
            "Defaults to: 'email'",
        )

    def handle(self, *args, campaign_types=None, **options):
        # used by campaign managers processes to queue messages to send
        self.queue = {
            key: mp.Queue(maxsize=app_settings.MAX_CONCURRENT_SENDERS)
            for key in CAMPAIGN_TYPE.keys()
        }

        # used to signals to sender they should exist once the queue has been emptied
        self.senders_quit_event = mp.Event()

        # used by email senders to make sure they're not going over the max rate
        self.rate_limiter = TokenBucket(
            max=app_settings.MAX_CONCURRENT_SENDERS * 2,
            rate=app_settings.MAX_SENDING_RATE,
        )

        # allow us to measure the sending speed
        self.rate_meter = RateMeter(0.3, 0.5)

        # used by the main process to monitor the sender processes
        self.sender_processes: Dict[str, List[mp.Process]] = {
            key: [] for key in CAMPAIGN_TYPE.keys()
        }

        # used by sender processes to signal when a campaign failed, so that it can be stopped by the main process
        self.sender_pipes: Dict[str, List[mp.connection.Connection]] = {
            key: [] for key in CAMPAIGN_TYPE.keys()
        }

        # used by the main process to monitor campaign managers and tell them to quit
        self.campaign_manager_processes: Dict[
            str, Dict[int, Tuple[mp.Process, mp.Event]]
        ] = {key: {} for key in CAMPAIGN_TYPE.keys()}

        self._setup_signals()
        self.run_loop(campaign_types)

    def _setup_signals(self):
        signal.signal(signal.SIGINT, signal.default_int_handler)
        signal.signal(signal.SIGTERM, gracefully_exit)
        signal.signal(signal.SIGUSR1, self.print_stats)
        signal.signal(signal.SIGUSR2, print_stack_trace)

    @contextlib.contextmanager
    def _setup_signal_handlers_for_children(self):
        try:
            signal.pthread_sigmask(signal.SIG_BLOCK, [signal.SIGTERM, signal.SIGINT])
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            signal.signal(signal.SIGUSR1, signal.SIG_DFL)
            signal.signal(signal.SIGUSR2, signal.SIG_DFL)
            yield
        finally:
            self._setup_signals()
            signal.pthread_sigmask(signal.SIG_UNBLOCK, [signal.SIGTERM, signal.SIGINT])

    def print_stats(self, sig, stack):
        for campaign_type in CAMPAIGN_TYPE.keys():
            values = {
                "campaign_type": campaign_type.upper(),
                "queue_size": self.queue[campaign_type].qsize(),
                "sender_processes": [
                    process.pid for process in self.sender_processes[campaign_type]
                ],
                "campaign_managers": [
                    process.pid
                    for process, _ in self.campaign_manager_processes[
                        campaign_type
                    ].values()
                ],
                "bucket_capacity": self.rate_limiter.peek(),
                "sending_rate": self.rate_meter.current_rate(),
            }
            self.stderr.write(self.STATS_MESSAGE % values, ending="\n\n")

    def start_sender_processes(self, campaign_type):
        sender_processes = self.sender_processes[campaign_type]
        queue = self.queue[campaign_type]

        for i in range(app_settings.MAX_CONCURRENT_SENDERS - len(sender_processes)):
            if i == 0:
                logger.info(f"\n{campaign_type.upper()}:")
            recv_conn, send_conn = mp.Pipe(duplex=False)
            process = mp.Process(
                target=CAMPAIGN_TYPE[campaign_type]["sender_process"],
                kwargs={
                    "queue": queue,
                    "error_channel": send_conn,
                    "rate_limiter": self.rate_limiter,
                    "rate_meter": self.rate_meter,
                    "quit_event": self.senders_quit_event,
                },
            )
            process.daemon = True

            sender_processes.append(process)
            self.sender_pipes[campaign_type].append(recv_conn)
            # let's close SQL connection to make sure it is not shared with children
            connection.close()
            with self._setup_signal_handlers_for_children():
                process.start()
            logger.info(
                _(
                    "Started sender process %(process_pid)s for %(campaign_type)s campaigns"
                )
                % {"process_pid": process.pid, "campaign_type": campaign_type}
            )

    def check_campaigns(self, campaign_type):
        CampaignModel = CAMPAIGN_TYPE[campaign_type]["CampaignModel"]
        campaigns = CampaignModel.objects.outbox()
        campaign_manager_process = self.campaign_manager_processes[campaign_type]
        queue = self.queue[campaign_type]

        for campaign in campaigns:
            if (
                campaign.status == CampaignModel.STATUS_WAITING
                and campaign.id in campaign_manager_process
            ):
                # we need to cancel that task
                logger.info(
                    _(
                        "Stopping campaign manager n°%(campaign_id)s (%(campaign_name)s)..."
                    )
                    % {"campaign_id": campaign.id, "campaign_name": campaign.name[20:]}
                )
                _process, quit_event = campaign_manager_process[campaign.id]
                quit_event.set()

            if (
                campaign.status == CampaignModel.STATUS_SENDING
                and campaign.id not in campaign_manager_process
            ):
                quit_event = mp.Event()
                process = mp.Process(
                    target=CAMPAIGN_TYPE[campaign_type]["manager_process"],
                    kwargs={
                        "campaign": campaign,
                        "queue": queue,
                        "quit_event": quit_event,
                    },
                )
                process.daemon = True
                campaign_manager_process[campaign.id] = (process, quit_event)
                # let's close SQL connection to make sure it is not shared with children
                connection.close()
                with self._setup_signal_handlers_for_children():
                    process.start()
                logger.info(
                    _(
                        "Started %(campaign_type)s campaign manager %(process_pid)s for %(campaign)s"
                    )
                    % {
                        "campaign_type": campaign_type,
                        "process_pid": process.pid,
                        "campaign": repr(campaign),
                    }
                )

    def monitor_processes(self, campaign_type):
        CampaignModel = CAMPAIGN_TYPE[campaign_type]["CampaignModel"]
        sender_processes = self.sender_processes[campaign_type]
        campaign_manager_process = self.campaign_manager_processes[campaign_type]
        sender_pipes = self.sender_pipes[campaign_type]

        sender_sentinels = [p.sentinel for p in sender_processes]
        campaign_manager_sentinels = {
            p.sentinel: c_id for c_id, (p, _e) in campaign_manager_process.items()
        }

        events = mpc.wait(
            sender_sentinels + list(campaign_manager_sentinels) + sender_pipes,
            timeout=app_settings.POLLING_INTERVAL,
        )

        stopped_senders = []
        stopped_campaign_managers = []
        campaign_errors = set()

        for event in events:
            if event in sender_sentinels:
                stopped_senders.append(sender_sentinels.index(event))
            if event in campaign_manager_sentinels:
                stopped_campaign_managers.append(campaign_manager_sentinels[event])
            elif event in sender_pipes:
                while event.poll():
                    try:
                        campaign_errors.add(event.recv())
                    except EOFError:
                        break

        # reverse=True is important as we're removing elements from the sender_processes list by index
        for i in sorted(stopped_senders, reverse=True):
            process = sender_processes[i]
            # let's reap process to avoid zombies
            process.join()

            logger.error(
                _("Sender process %(process_pid)s unexpectedly quit...")
                % {"process_pid": process.pid}
            )
            del sender_processes[i]
            del sender_pipes[i]

        for campaign_id in sorted(stopped_campaign_managers):
            process, _quit_event = campaign_manager_process[campaign_id]
            pid = process.pid
            # let's reap process to avoid zombies
            process.join()

            try:
                campaign = CampaignModel.objects.get(id=campaign_id)
            except CampaignModel.DoesNotExist:
                # the campaign has most likely be deleted
                campaign = None

            if campaign and campaign.status != CampaignModel.STATUS_SENDING:
                # process was asked to stop and did so correctly
                logger.info(
                    _(
                        "Campaign manager %(pid)s correctly stopped... Was taking care of %(campaign)s."
                    )
                    % {"pid": pid, "campaign": repr(campaign)},
                    exc_info=True,
                )
            else:
                logger.error(
                    _(
                        "Campaign manager %(pid)s abruptly stops. Was taking care of %(campaign)s."
                    )
                    % {"pid": pid, "campaign": repr(campaign)},
                    exc_info=True,
                )

            del campaign_manager_process[campaign_id]

        for campaign_id in campaign_errors:
            CampaignModel.objects.filter(id=campaign_id).update(
                status=CampaignModel.STATUS_ERROR
            )
            if campaign_id in campaign_manager_process:
                logger.error(
                    _(
                        "Unexpected error while trying to send message from campaign %(campaign_id)s...\n"
                    )
                    % {"campaign_id": campaign_id}
                )
                campaign_manager_process[campaign_id][1].set()

    def run_loop(self, campaign_types):
        try:
            while True:
                for campaign_type in campaign_types:
                    self.start_sender_processes(campaign_type)
                    self.check_campaigns(campaign_type)
                    self.monitor_processes(campaign_type)

        except (GracefulExit, KeyboardInterrupt):
            logger.info(_("Asked to quit, asking all subprocesses to exit..."))
            self.senders_quit_event.set()

            for campaign_type in campaign_types:
                for _proces, event in self.campaign_manager_processes[
                    campaign_type
                ].values():
                    event.set()

            logger.info(_("Waiting for all subprocesses to gracefully exit..."))
            # active_children joins children so removes zombies
            for campaign_type in campaign_types:
                while mp.active_children():
                    sender_processes = sum(self.sender_processes.values(), [])
                    mpc.wait(
                        [p.sentinel for p in sender_processes]
                        + [
                            p.sentinel
                            for p, _e in self.campaign_manager_processes[
                                campaign_type
                            ].values()
                        ]
                    )

            logger.info(_("All subprocesses have exited!"))



# Generated by Django 2.2.23 on 2021-11-09 08:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0021_create_push_campaign_models")]

    operations = [
        migrations.AddField(
            model_name="campaign",
            name="end_date",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Campaign end date"
            ),
        ),
        migrations.AddField(
            model_name="campaign",
            name="start_date",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Campaign start date"
            ),
        ),
        migrations.AddField(
            model_name="pushcampaign",
            name="end_date",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Campaign end date"
            ),
        ),
        migrations.AddField(
            model_name="pushcampaign",
            name="start_date",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Campaign start date"
            ),
        ),
    ]



import contextlib
import logging
import multiprocessing as mp
import signal
import time
import traceback
from ctypes import c_double, c_ulong
from queue import Empty, Full

logger = logging.getLogger(__name__)


class GracefulExit(Exception):
    pass


def _current_time():
    """Returns monotonic current time in seconds.

    Splitted off in function to increase testability.

    :return: current time in seconds, guaranteed monotonic
    """
    return time.monotonic()


def gracefully_exit(sig, stack):
    """
    Signal handler that raises a GracefulExit exception

    :raise: :class:`nuntius.synchronize.GracefulExit`
    """
    raise GracefulExit()


def print_stack_trace(sig, stack):
    """
    Signal handler that prints the current track before resuming process
    """
    traceback.print_stack(stack)


def unexpected_exc_logger(proc):
    """
    Decorator that intercepts exception and prints them
    :param proc:
    :return:
    """

    @contextlib.wraps(proc)
    def wrapper(*args, **kwargs):
        try:
            return proc(*args, **kwargs)
        except Exception:
            logger.exception("Unexpected error.")
            raise

    return wrapper


def reset_sigmask(proc):
    """
    Decorator that sets an empty sigmask when the function is called
    """

    @contextlib.wraps(proc)
    def wrapper(*args, **kwargs):
        signal.pthread_sigmask(signal.SIG_SETMASK, [])
        return proc(*args, **kwargs)

    return wrapper


class RateLimiter:
    def take(self):
        return


class TokenBucket(RateLimiter):
    """
    Simple multiprocessing token bucket implementation of the RateLimiter Interface

    Token buckets have the following principles :
    - They fill up at a fixed rate
    - They have a maximum capacity and will stop filling up when it is reached
    - Whenever a process calls `take`, if the bucket is full enough is is decreased ;
      if it is not the case, the process (and others) are blocked until is has filled
      in enough.
    """

    def __init__(self, max: int, rate: float):
        """Create a new TokenBucket.

        :param max: The maximum number of tokens that may be stored in the bucket
        :type max: class:`int`
        :param rate: The rate at which the bucket fills in, in number of tokens per second
        :type rate: class:`float`
        """
        self.max = max
        self.rate = rate
        self._var_lock = mp.RLock()
        self._wait_lock = mp.RLock()
        self._timestamp = mp.Value(c_double, lock=False)
        self._timestamp.value = time.monotonic()
        self._capacity = mp.Value(c_double, lock=False)
        self._capacity.value = self.max

    def _update(self):
        with self._var_lock:
            now = _current_time()
            self._capacity.value = min(
                self._capacity.value + self.rate * (now - self._timestamp.value),
                self.max,
            )
            self._timestamp.value = now

    def take(self, n=1):
        """
        Try to take a token, or wait for the bucket to fill in enough.

        If several processes take at the same time, the first one to call will takes
        a lock and will be guaranteed to be unblocked first. No such guarantee exists
        for the other processes.
        """
        with self._wait_lock:
            with self._var_lock:
                self._update()
                self._capacity.value -= n
                capacity = self._capacity.value

            if capacity < 0:
                time.sleep(-self._capacity.value / self.rate)

    def peek(self):
        with self._var_lock:
            self._update()
            return self._capacity.value


class RateMeter:
    def __init__(self, alpha: float, window: float):
        self._alpha = alpha
        self._window = window
        self._lock = mp.RLock()
        self._current_rate = mp.Value(c_double, 0, lock=False)
        self._last_window = mp.Value(
            c_ulong, int(_current_time() / self._window), lock=False
        )
        self._current_counter = mp.Value(c_ulong, 0, lock=False)

    def _update(self):
        with self._lock:
            current_window = int(_current_time() / self._window)
            if current_window > self._last_window.value:
                time_diff = current_window - self._last_window.value
                alpha, beta = self._alpha, 1 - self._alpha
                self._current_rate.value = beta ** (time_diff - 1) * (
                    beta * self._current_rate.value
                    + alpha * (self._current_counter.value / self._window)
                )
                self._current_counter.value = 0
                self._last_window.value = current_window

    def count_up(self, n=1):
        with self._lock:
            self._update()
            self._current_counter.value += n

    def current_rate(self):
        with self._lock:
            return self._current_rate.value


def get_from_queue_or_quit(queue: mp.Queue, event: mp.Event, polling_period: float):
    while True:
        if event.is_set():
            raise GracefulExit()
        try:
            return queue.get(timeout=polling_period)
        except Empty:
            pass


def put_in_queue_or_quit(
    queue: mp.Queue, value, event: mp.Event, polling_period: float
):
    while True:
        if event.is_set():
            raise GracefulExit()
        try:
            return queue.put(value, timeout=polling_period)
        except Full:
            pass



from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import fields
from django.utils.translation import gettext_lazy as _


class BaseSegment:
    def get_display_name(self):
        raise NotImplementedError()

    def get_subscribers_queryset(self):
        raise NotImplementedError()

    def get_subscribers_count(self):
        raise NotImplementedError

    class Meta:
        swappable = "NUNTIUS_SEGMENT_MODEL"
        verbose_name = _("Segment")
        verbose_name_plural = _("Segments")


class BaseSubscriberManager(models.Manager):
    def set_subscriber_status(self, email_address, status):
        try:
            subscriber = self.get(email=email_address)
        except ObjectDoesNotExist:
            return
        subscriber.subscriber_status = status
        subscriber.save(update_fields=["subscriber_status"])

    def get_subscriber(self, email_address):
        return self.filter(email=email_address).last()


class AbstractSubscriber:
    STATUS_SUBSCRIBED = 1
    STATUS_UNSUBSCRIBED = 2
    STATUS_BOUNCED = 3
    STATUS_COMPLAINED = 4
    STATUS_CHOICES = (
        (STATUS_SUBSCRIBED, _("Subscribed")),
        (STATUS_UNSUBSCRIBED, _("Unsubscribed")),
        (STATUS_BOUNCED, _("Bounced")),
        (STATUS_COMPLAINED, _("Complained")),
    )

    def get_subscriber_status(self):
        if hasattr(self, "subscriber_status"):
            return self.subscriber_status
        raise NotImplementedError()

    def get_subscriber_email(self):
        if hasattr(self, "email"):
            return self.email

        raise NotImplementedError()

    def get_subscriber_push_devices(self):
        if hasattr(self, "push_devices"):
            return self.push_devices

        return []

    def get_subscriber_data(self):
        return {"email": self.get_subscriber_email()}

    class Meta:
        abstract = True
        verbose_name = _("Subscriber")
        verbose_name_plural = _("Subscribers")


class BaseSubscriber(AbstractSubscriber, models.Model):
    objects = BaseSubscriberManager()

    email = fields.EmailField(max_length=255, verbose_name=_("Email"))
    subscriber_status = fields.IntegerField(
        choices=AbstractSubscriber.STATUS_CHOICES, verbose_name=_("Subscriber status")
    )

    class Meta(AbstractSubscriber.Meta):
        swappable = "NUNTIUS_SUBSCRIBER_MODEL"



import re
from secrets import token_urlsafe

from django.db import models
from django.db.models import fields, Sum, Value
from django.db.models.functions import Coalesce
from django.template import Template
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from stdimage import StdImageField

from nuntius import app_settings
from nuntius.models.mixins import AbstractCampaign
from nuntius.utils.messages import generate_plain_text

MOSAICO_TO_DJANGO_TEMPLATE_VARS = re.compile(r"\[([A-Z_-]+)]")


class Campaign(AbstractCampaign):
    message_from_name = fields.CharField(_('"From" name'), max_length=255, blank=True)
    message_from_email = fields.EmailField(_('"From" email address'), max_length=255)
    message_reply_to_name = fields.CharField(
        _('"Reply to" name'), max_length=255, blank=True
    )
    message_reply_to_email = fields.CharField(
        _('"Reply to" email address'), max_length=255, blank=True
    )
    message_subject = fields.CharField(
        _("Message subject line"), max_length=255, blank=True
    )
    message_mosaico_data = fields.TextField(_("Mosaico data"), blank=True)
    message_content_html = fields.TextField(_("Message content (HTML)"), blank=True)
    message_content_text = fields.TextField(_("Message content (text)"), blank=True)

    def save(self, *args, **kwargs):
        if self.message_mosaico_data:
            self.message_content_text = generate_plain_text(self.message_content_html)
        super().save(*args, **kwargs)

    def get_sent_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .exclude(result=CampaignSentStatusType.PENDING)
            .count()
        )

    def get_ok_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(result=CampaignSentStatusType.OK)
            .count()
        )

    def get_bounced_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(result=CampaignSentStatusType.BOUNCED)
            .count()
        )

    def get_complained_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(result=CampaignSentStatusType.COMPLAINED)
            .count()
        )

    def get_blocked_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(result=CampaignSentStatusType.BLOCKED)
            .count()
        )

    def get_open_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).aggregate(
            count=Coalesce(Sum("open_count"), Value(0))
        )["count"]

    def get_unique_open_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(open_count__gt=0)
            .count()
        )

    def get_click_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).aggregate(
            count=Coalesce(Sum("click_count"), Value(0))
        )["count"]

    def get_unique_click_count(self):
        return (
            CampaignSentEvent.objects.filter(campaign=self)
            .filter(click_count__gt=0)
            .count()
        )

    def get_event_for_subscriber(self, subscriber):
        event, _ = CampaignSentEvent.objects.get_or_create(
            campaign=self,
            subscriber=subscriber,
            defaults={"email": subscriber.get_subscriber_email()},
        )
        return event

    def __repr__(self):
        return f"Campaign(id={self.id!r}, name={self.name!r})"

    @cached_property
    def html_template(self):
        return Template(self.message_content_html)

    @cached_property
    def text_template(self):
        return Template(self.message_content_text)

    @property
    def from_header(self):
        if self.message_from_name:
            return f"{self.message_from_name} <{self.message_from_email}>"
        return self.message_from_email

    @property
    def reply_to_header(self):
        if self.message_reply_to_email:
            if self.message_reply_to_name:
                return f"{self.message_reply_to_name} <{self.message_reply_to_email}>"
            return self.message_reply_to_email
        return None

    class Meta:
        verbose_name = _("email campaign")
        verbose_name_plural = _("email campaigns")


class CampaignSentStatusType:
    PENDING = "P"
    UNKNOWN = "?"
    REJECTED = "RE"
    OK = "OK"
    BOUNCED = "BC"
    COMPLAINED = "C"
    UNSUBSCRIBED = "U"
    BLOCKED = "BL"
    ERROR = "E"

    CHOICES = (
        (PENDING, _("Sending")),
        (UNKNOWN, _("Unknown")),
        (REJECTED, _("Rejected by server")),
        (OK, _("Sent")),
        (BOUNCED, _("Bounced")),
        (COMPLAINED, _("Complained")),
        (UNSUBSCRIBED, _("Unsubscribed")),
        (BLOCKED, _("Blocked temporarily")),
        (ERROR, _("Error")),
    )


class CampaignSentEvent(models.Model):
    subscriber = models.ForeignKey(
        app_settings.NUNTIUS_SUBSCRIBER_MODEL,
        models.SET_NULL,
        verbose_name=_("Subscriber"),
        null=True,
        blank=True,
    )
    email = models.EmailField(_("Email address at sending time"))
    campaign = models.ForeignKey(
        "Campaign", models.CASCADE, verbose_name=_("Campaign"), null=True, blank=True
    )
    datetime = models.DateTimeField(_("Sending time"), auto_now_add=True)
    result = models.CharField(
        _("Operation result"),
        max_length=2,
        default=CampaignSentStatusType.PENDING,
        choices=CampaignSentStatusType.CHOICES,
    )
    esp_message_id = models.CharField(
        _("ID given by the sending server"),
        unique=True,
        max_length=255,
        null=True,
        editable=False,
    )

    def generate_tracking_id():
        return token_urlsafe(9)

    tracking_id = models.CharField(
        max_length=12,
        default=generate_tracking_id,
        null=True,
        editable=False,
        unique=True,
    )
    open_count = models.IntegerField(_("Open count"), default=0, editable=False)
    click_count = models.IntegerField(_("Click count"), default=0, editable=False)

    class Meta:
        unique_together = ("campaign", "subscriber")
        verbose_name = _("email sent event")
        verbose_name_plural = _("email sent events")
        # this (email, datetime) index is required to handle bouncing rules
        # see func:`nuntius.actions.update_subscriber`
        indexes = [
            models.Index(fields=["email", "datetime"]),
            models.Index(fields=["subscriber", "datetime"]),
        ]
        ordering = ["-datetime"]


class MosaicoImage(models.Model):
    file = StdImageField(upload_to="mosaico", variations={"thumbnail": (90, 90)})
    created = fields.DateTimeField(auto_now_add=True)



from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse, path
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from nuntius import app_settings
from nuntius.admin.panels import subscriber_class
from nuntius.models import Campaign
from nuntius.models.push_campaigns import PushCampaign, PushCampaignSentEvent


class PushCampaignAdmin(admin.ModelAdmin):
    search_fields = ("name", "notification_title")
    autocomplete_fields = ("segment",)
    prepopulated_fields = {"utm_name": ("name",)}
    fieldsets = (
        (None, {"fields": ("created", "updated", "name", "utm_name")}),
        (
            _("Notification"),
            {
                "fields": (
                    "notification_title",
                    "notification_url",
                    "notification_body",
                    "notification_tag",
                )
            },
        ),
        (
            _("Sending details"),
            {
                "fields": (
                    "start_date",
                    "end_date",
                    "first_sent",
                    "segment",
                    "segment_subscribers",
                    "status",
                    "send_button",
                )
            },
        ),
        (
            _("Sending reports"),
            {"fields": ("sent_to", "sent_ok", "sent_ko", "click_count")},
        ),
    )
    list_display = (
        "name",
        "notification_title",
        "segment",
        "status",
        "send_button",
        "sent_to",
    )
    list_filter = ("status",)
    readonly_fields = (
        "created",
        "updated",
        "first_sent",
        "segment_subscribers",
        "status",
        "send_button",
        "sent_to",
        "sent_ok",
        "sent_ko",
        "click_count",
    )
    save_as = True

    def segment_subscribers(self, instance):
        if instance.segment is None:
            return subscriber_class().objects.count()
        return instance.segment.get_subscribers_count()

    segment_subscribers.short_description = _("Subscribers")

    def sent_to(self, instance):
        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:nuntius_pushcampaignsentevent_changelist")
            + "?campaign_id__exact="
            + str(instance.pk),
            str(instance.get_sent_count()),
        )

    sent_to.short_description = _("Sent to")

    def sent_ok(self, instance):
        return instance.get_ok_count()

    sent_ok.short_description = _("OK")

    def sent_ko(self, instance):
        return instance.get_ko_count()

    sent_ko.short_description = _("KO")

    def click_count(self, instance):
        return instance.get_click_count()

    click_count.short_description = _("Click count")

    def send_button(self, instance):
        if instance.pk is None:
            return mark_safe("-")
        if instance.status == Campaign.STATUS_SENDING:
            return format_html(
                '<a href="{}" class="button">' + _("Pause") + "</a>",
                reverse("admin:nuntius_pushcampaign_pause", args=[instance.pk]),
            )
        return format_html(
            '<a href="{}" class="button">' + _("Send") + "</a>",
            reverse("admin:nuntius_pushcampaign_send", args=[instance.pk]),
        )

    send_button.short_description = _("Send")

    def get_urls(self):
        return [
            path(
                "<pk>/send/",
                self.admin_site.admin_view(self.send_view),
                name="nuntius_pushcampaign_send",
            ),
            path(
                "<pk>/pause/",
                self.admin_site.admin_view(self.pause_view),
                name="nuntius_pushcampaign_pause",
            ),
        ] + super().get_urls()

    def send_view(self, request, pk):
        campaign = PushCampaign.objects.get(pk=pk)
        campaign.status = PushCampaign.STATUS_SENDING
        campaign.save(update_fields=["status"])

        return redirect(reverse("admin:nuntius_pushcampaign_change", args=[pk]))

    def pause_view(self, request, pk):
        campaign = PushCampaign.objects.get(pk=pk)

        campaign.status = PushCampaign.STATUS_WAITING
        campaign.save(update_fields=["status"])

        return redirect(reverse("admin:nuntius_pushcampaign_change", args=[pk]))


class TrackingFilter(admin.SimpleListFilter):
    title = _("Clicked")
    parameter_name = "clicked"

    def lookups(self, request, model_admin):
        return (("1", _("Yes")), ("0", _("No")))

    def queryset(self, request, queryset):
        if self.value() == "1":
            return queryset.exclude(click_count=0)
        if self.value() == "0":
            return queryset.filter(click_count=0)
        return queryset


class PushCampaignSentEventAdmin(admin.ModelAdmin):
    def has_change_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False

    actions = None
    readonly_fields = ("subscriber_filter", "campaign_filter")
    list_filter = ("result", TrackingFilter)
    list_display_links = None

    def get_list_display(self, request):
        list_display = ("datetime", "result", "click_count")

        if request.GET.get("campaign_id__exact") is None:
            list_display = ("campaign_filter", *list_display)
        else:
            list_display = ("campaign", *list_display)

        if request.GET.get("subscriber_id__exact") is None:
            list_display = ("subscriber_filter", *list_display)
        else:
            list_display = ("subscriber", *list_display)

        return list_display

    def subscriber_filter(self, instance):
        if instance.subscriber is None:
            return "-"

        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:nuntius_pushcampaignsentevent_changelist")
            + "?subscriber_id__exact="
            + str(instance.subscriber_id),
            str(instance.subscriber),
        )

    subscriber_filter.short_description = _("Subscriber")

    def campaign_filter(self, instance):
        if instance.campaign is None:
            return "-"

        return format_html(
            "<a href={}>{}</a>",
            reverse("admin:nuntius_pushcampaignsentevent_changelist")
            + "?campaign_id__exact="
            + str(instance.campaign_id),
            str(instance.campaign),
        )

    campaign_filter.short_description = _("Campaign")

    def changelist_view(self, request, extra_context=None):
        title = _("Sent events")
        campaign, subscriber = (None, None)

        if request.GET.get("campaign_id__exact") is not None:
            campaign = PushCampaign.objects.filter(
                id=request.GET.get("campaign_id__exact")
            ).first()
        if request.GET.get("subscriber_id__exact") is not None:
            subscriber = (
                subscriber_class()
                .objects.filter(id=request.GET.get("subscriber_id__exact"))
                .first()
            )

        if campaign and subscriber:
            title = _(
                f"Sent event for campaign %(campaign)s and subscriber %(subscriber)s"
            ) % {"campaign": str(campaign), "subscriber": str(subscriber)}
        elif campaign:
            title = mark_safe(
                _("Sent events for campaign %s")
                % (
                    format_html(
                        '<a href="{}">{}</a>',
                        reverse(
                            "admin:nuntius_pushcampaign_change", args=[campaign.pk]
                        ),
                        str(campaign),
                    ),
                )
            )
        elif subscriber:
            title = _("Sent events for subscriber %s") % (str(subscriber),)

        return super().changelist_view(request, extra_context={"title": title})


if not app_settings.DISABLE_DEFAULT_ADMIN:
    admin.site.register(PushCampaign, PushCampaignAdmin)
    admin.site.register(PushCampaignSentEvent, PushCampaignSentEventAdmin)
